#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#-------------------------------------------------------------------------------------------------

import re
import os
import logging
from shutil import copyfile

from hostattachment.platform.linux import definitions as linux_hak_defs
from hostattachment.platform.linux import messages
from hostattachment.platform.linux.common import generate_backup_filename
from hostattachment.exceptions import HostAttachmentError
from hostattachment.platform.linux.classes import get_os_string
from hostattachment.classes import AttachmentConfiguration

from xiv_xsf.host.os.linux.udev import udev_trigger, UdevError

UDEV_RULES_PATH = os.path.join(os.path.sep, "etc", "udev", "rules.d")

RHEL5_MP_RULES_FILE="/etc/udev/rules.d/40-multipath.rules"

class UdevRuleNotFound(Exception):
    pass

def read_file(filename):
    with file(filename) as fh:
        lines = fh.readlines()

    return lines

def write_file(filename, lines):
    with file(filename, 'w') as fh:
        fh.writelines(lines)

def verify_no_last_rule(filename):
    if not os.path.exists(filename):
        return False

    for line in read_file(filename):
        line = line.strip()
        if 'OPTIONS="last_rule"' in line and not line.startswith("#"):
            logging.warning(messages.RHEL5_RULES_FILE_HAS_LAST_RULE_OPTION)
            return False

    return True

def remove_last_rule_option(lines):
    """Removes the OPTIONS="last_rule" thing from the given rulefile content.
    Needed for updating the 40-multipath.rules file in RHEL5"""
    newlines = []
    for line in lines:
        line = line.replace('OPTIONS="last_rule"', '')
        line = re.sub(',[ \t]*$', '', line)
        newlines.append(line)

    return newlines

class UdevRules(object):
    def __init__(self):
        self.os_string = get_os_string()
        self.rules_source = linux_hak_defs.UDEV_RULES_SOURCE
        self.udev_rules = []
        self.udev_rules.append('44-ibm-xiv.rules')
        self.udev_rules.append('46-ibm-xiv.rules')
        attachment_configuration = AttachmentConfiguration.getConfiguration()
        self.backup_path = attachment_configuration.backup_path

    def is_configured(self):
        # For RHEL5, make sure that RHEL5_MP_RULES_FILE does not contain last_rule option
        no_last_rule = True
        if self.os_string == "redhat5":
            no_last_rule = verify_no_last_rule(RHEL5_MP_RULES_FILE)

        # Test XIV rules
        xiv_rules = self.dst_rules_configured()

        return no_last_rule and xiv_rules

    def _get_src_file(self, dir, filename):
        """Tries all optional filenames in an array and returns the first
        match. In this case, try os-specific first and then generic version
        """
        options = []
        options.append('%s/%s.%s' % (dir, filename, self.os_string))
        options.append('%s/%s' % (dir, filename))

        for option in options:
            if (os.path.exists(option)):
                return option

        raise UdevRuleNotFound(messages.CANT_FIND_ANY_FILE % (options) )

    def rules_differ(self, rules1, rules2):
        """Compares the two rules file contents"""
        # TODO should be a smarter compare (i.e. ignore comments and spaces)
        if rules1 != rules2:
            return True

        return False

    def dst_rules_configured(self):
        for filename in self.udev_rules:
            try:
                src_filename = self._get_src_file(self.rules_source, filename)
            except UdevRuleNotFound:
                continue # Probably no such rule for this OS
            src = file(src_filename)
            try:
                dst = file('%s/%s' % (UDEV_RULES_PATH, filename))
            except IOError:
                logging.error(messages.UDEV_RULES_FILE_DOESNT_EXIST % (filename))
                return False

            if self.rules_differ(src.read(), dst.read()):
                logging.error(messages.UDEV_RULE_NOT_CONFIGURED % (filename))
                return False

        return True

    def backup(self, src):
        if (os.path.exists(src)):
            backup_file = generate_backup_filename(self.backup_path,
                                            os.path.basename(src))
            logging.info(messages.BACKING_UP_SRC_TO_DST % (src, backup_file))
            copyfile(src, backup_file)

    def fix_40_rules_file_rhel5(self):
        """Removes lines with 'last_rule' option inside"""
        if not os.path.exists(RHEL5_MP_RULES_FILE):
            return

        try:
            lines = read_file(RHEL5_MP_RULES_FILE)
        except Exception:
            raise HostAttachmentError("Error reading '%s'" % (RHEL5_MP_RULES_FILE))

        if verify_no_last_rule(RHEL5_MP_RULES_FILE):
            return # No need to make changes changes

        lines = remove_last_rule_option(lines)
        try:
            self.backup(RHEL5_MP_RULES_FILE)
        except Exception:
            logging.error(messages.BACKUP_FAILED % (RHEL5_MP_RULES_FILE))
            raise
        else:
            logging.info(messages.WRITING_RHEL5_RULES_FILE % (RHEL5_MP_RULES_FILE))
            write_file(RHEL5_MP_RULES_FILE, lines)

    def configure(self):
        for basename in self.udev_rules:
            try:
                src_filename = self._get_src_file(self.rules_source, basename)
            except UdevRuleNotFound:
                continue # Probably no such rule for this OS

            dst_filename = os.path.join(UDEV_RULES_PATH, basename)
            self.backup(dst_filename)
            try:
                copyfile(src_filename, dst_filename)
                logging.info(messages.UDEV_RULE_CONFIG_SUCCESS % (basename))
            except IOError:
                raise HostAttachmentError(messages.UDEV_RULE_CONFIG_FAILURE % (basename) )
        if self.os_string == "redhat5":
            self.fix_40_rules_file_rhel5()

        try:
            udev_trigger()
        except UdevError:
            # This is usually okay
            logging.warning(messages.UDEV_TRIGGER_FAILED)
        else:
            logging.info(messages.UDEV_TRIGGERED)
